import "dart:io";

import "package:moxxyv2/ui/constants.dart";
import "package:moxxyv2/ui/helpers.dart";
import "package:moxxyv2/ui/widgets/topbar.dart";
import "package:moxxyv2/ui/widgets/avatar.dart";
import "package:moxxyv2/ui/widgets/textfield.dart";
import "package:moxxyv2/ui/widgets/chat/media/media.dart";
import "package:moxxyv2/ui/widgets/chat/chatbubble.dart";
import "package:moxxyv2/ui/widgets/chat/typing.dart";
import "package:moxxyv2/ui/bloc/conversation_bloc.dart";
import "package:moxxyv2/ui/bloc/conversations_bloc.dart";
import "package:moxxyv2/ui/bloc/profile_bloc.dart";
import "package:moxxyv2/shared/models/message.dart";
import "package:moxxyv2/xmpp/xeps/xep_0085.dart";

import "package:flutter/material.dart";
import "package:flutter_bloc/flutter_bloc.dart";
import "package:flutter_speed_dial/flutter_speed_dial.dart";
import "package:swipeable_tile/swipeable_tile.dart";
import "package:flutter_vibrate/flutter_vibrate.dart";
import "package:get_it/get_it.dart";

enum ConversationOption {
  close,
  block
}

enum EncryptionOption {
  omemo,
  none
}

PopupMenuItem popupItemWithIcon(dynamic value, String text, IconData icon) {
  return PopupMenuItem(
    value: value,
    child: Row(
      children: [
        Padding(
          padding: const EdgeInsets.only(right: 8.0),
          child: Icon(icon)
        ),
        Text(text)
      ]
    )
  );
}

/// A custom version of the Topbar NameAndAvatar style to integrate with
/// bloc.
class _ConversationTopbarWidget extends StatelessWidget {
  bool _shouldRebuild(ConversationState prev, ConversationState next) {
    return prev.conversation?.title != next.conversation?.title
      || prev.conversation?.avatarUrl != next.conversation?.avatarUrl
      || prev.conversation?.chatState != next.conversation?.chatState;
  }

  Widget _buildChatState(ChatState state) {
    switch (state) {
      case ChatState.paused:
      case ChatState.active: {
        return const Text(
          "Online",
          style: TextStyle(
            color: Colors.green
          )
        );
      }
      case ChatState.composing: {
        // TODO: Colors
        return const TypingIndicatorWidget(Colors.black, Colors.white);
      }
      default: {
        return Container();
      }
    }
  }
  
  @override
  Widget build(BuildContext context) {
    return BlocBuilder<ConversationBloc, ConversationState>(
      buildWhen: (prev, next) => _shouldRebuild(prev, next),
      builder: (context, state) {
        return TopbarAvatarAndName(
          IntrinsicHeight(
            child: Column(
              children: [
                TopbarTitleText(state.conversation!.title),
                _buildChatState(state.conversation!.chatState)
              ]
            )
          ),
          AvatarWrapper(
            radius: 25.0,
            avatarUrl: state.conversation!.avatarUrl,
            altText: state.conversation!.title,
          ),
          () => GetIt.I.get<ProfileBloc>().add(
            ProfilePageRequestedEvent(
              false,
              conversation: state.conversation!
            )
          ),
          showBackButton: true,
          extra: [
            PopupMenuButton(
              onSelected: (result) {
                if (result == EncryptionOption.omemo) {
                  showNotImplementedDialog("End-to-End encryption", context);
                }
              },
              icon: const Icon(Icons.lock_open),
              itemBuilder: (BuildContext c) => [
                popupItemWithIcon(EncryptionOption.none, "Unencrypted", Icons.lock_open),
                popupItemWithIcon(EncryptionOption.omemo, "Encrypted", Icons.lock),
              ]
            ),
            PopupMenuButton(
              onSelected: (result) {
                switch (result) {
                  case ConversationOption.close: {
                    showConfirmationDialog(
                      "Close Chat",
                      "Are you sure you want to close this chat?",
                      context,
                      () {
                        context.read<ConversationsBloc>().add(
                          ConversationClosedEvent(state.conversation!.jid)
                        );
                        Navigator.of(context).pop();
                      }
                    );
                  }
                  break;
                  case ConversationOption.block: {
                    //_block(state, context);
                  }
                  break;
                }
              },
              icon: const Icon(Icons.more_vert),
              itemBuilder: (BuildContext c) => [
                popupItemWithIcon(ConversationOption.close, "Close chat", Icons.close),
                popupItemWithIcon(ConversationOption.block, "Block contact", Icons.block)
              ]
            )
          ]
        );
      }
    );
  }
}

class _ConversationBottomRow extends StatelessWidget {
  final TextEditingController controller;
  final ValueNotifier<bool> isSpeedDialOpen;

  const _ConversationBottomRow(this.controller, this.isSpeedDialOpen);
  
  @override
  Widget build(BuildContext context) {
    return BlocBuilder<ConversationBloc, ConversationState>(
      buildWhen: (prev, next) => prev.showSendButton != next.showSendButton,
      builder: (context, state) => Container(
        color: Theme.of(context).backgroundColor,
        child: Padding(
          padding: const EdgeInsets.all(8.0),
          child: Row(
            children: [
              Expanded(
                child: CustomTextField(
                  maxLines: 5,
                  minLines: 1,
                  hintText: "Send a message...",
                  isDense: true,
                  onChanged: (value) {
                    context.read<ConversationBloc>().add(
                      MessageTextChangedEvent(value)
                    );
                  },
                  contentPadding: textfieldPaddingConversation,
                  cornerRadius: textfieldRadiusConversation,
                  controller: controller,
                  topWidget: state.quotedMessage != null ? buildQuoteMessageWidget(
                    state.quotedMessage!,
                    resetQuote: () => context.read<ConversationBloc>().add(QuoteRemovedEvent())
                  ) : null
                )
              ),
              Padding(
                padding: const EdgeInsets.only(left: 8.0),
                // NOTE: https://stackoverflow.com/a/52786741
                //       Thank you kind sir
                child: SizedBox(
                  height: 45.0,
                  width: 45.0,
                  child: FittedBox(
                    child: SpeedDial(
                      icon: state.showSendButton ? Icons.send : Icons.add,
                      visible: true,
                      curve: Curves.bounceInOut,
                      backgroundColor: primaryColor,
                      // TODO: Theme dependent?
                      foregroundColor: Colors.white,
                      openCloseDial: isSpeedDialOpen,
                      onPress: () {
                        if (state.showSendButton) {
                          context.read<ConversationBloc>().add(
                            MessageSentEvent()
                          );
                          controller.text = "";
                        } else {
                          isSpeedDialOpen.value = true;
                        }
                      },
                      children: [
                        SpeedDialChild(
                          child: const Icon(Icons.image),
                          onTap: () {
                            showNotImplementedDialog("sending files", context);
                            //Navigator.pushNamed(context, sendFilesRoute);
                          },
                          backgroundColor: primaryColor,
                          // TODO: Theme dependent?
                          foregroundColor: Colors.white,
                          label: "Send Image"
                        ),
                        SpeedDialChild(
                          child: const Icon(Icons.photo_camera),
                          onTap: () {
                            showNotImplementedDialog("sending files", context);
                          },
                          backgroundColor: primaryColor,
                          // TODO: Theme dependent?
                          foregroundColor: Colors.white,
                          label: "Take photo"
                        ),
                        SpeedDialChild(
                          child: const Icon(Icons.attach_file),
                          onTap: () {
                            showNotImplementedDialog("sending files", context);
                          },
                          backgroundColor: primaryColor,
                          // TODO: Theme dependent?
                          foregroundColor: Colors.white,
                          label: "Add file"
                        )
                      ]
                    )
                  )
                )
              )
            ]
          )
        )
      )
    );
  }
}

class ConversationPage extends StatefulWidget {
  const ConversationPage({ Key? key }) : super(key: key);

  @override
  _ConversationPageState createState() => _ConversationPageState();
}

class _ConversationPageState extends State<ConversationPage> {
  final TextEditingController _controller;
  final ValueNotifier<bool> _isSpeedDialOpen;

  _ConversationPageState() :
    _isSpeedDialOpen = ValueNotifier(false),
    _controller = TextEditingController(),
    super();
  
  @override
  void dispose() {
    _controller.dispose();

    super.dispose();
  }

  Widget _renderBubble(ConversationState state, BuildContext context, int _index, double maxWidth) {
    // TODO: Since we reverse the list: Fix start, end and between
    final index = state.messages.length - 1 - _index;
    Message item = state.messages[index];
    bool start = index - 1 < 0 ? true : state.messages[index - 1].sent != item.sent;
    bool end = index + 1 >= state.messages.length ? true : state.messages[index + 1].sent != item.sent;
    bool between = !start && !end;

    return SwipeableTile.swipeToTrigger(
      direction: SwipeDirection.horizontal,
      swipeThreshold: 0.2,
      onSwiped: (_) => context.read<ConversationBloc>().add(MessageQuotedEvent(item)),
      backgroundBuilder: (_, direction, progress) {
        // NOTE: Taken from https://github.com/watery-desert/swipeable_tile/blob/main/example/lib/main.dart#L240
        //       and modified.
        bool vibrated = false;
        return AnimatedBuilder(
          animation: progress,
          builder: (_, __) {
            if (progress.value > 0.9999 && !vibrated) {
              Vibrate.feedback(FeedbackType.light);
              vibrated = true;
            } else if (progress.value < 0.9999) {
              vibrated = false;
            }

            return Container(
              alignment: direction == SwipeDirection.endToStart ? Alignment.centerRight : Alignment.centerLeft,
              child: Padding(
                padding: EdgeInsets.only(
                  right: direction == SwipeDirection.endToStart ? 24.0 : 0.0,
                  left: direction == SwipeDirection.startToEnd ? 24.0 : 0.0
                ),
                child: Transform.scale(
                  scale: Tween<double>(
                    begin: 0.0,
                    end: 1.2,
                  )
                  .animate(
                    CurvedAnimation(
                      parent: progress,
                      curve: const Interval(0.5, 1.0,
                        curve: Curves.linear),
                    ),
                  )
                  .value,
                  child: Container(
                    decoration: BoxDecoration(
                      color: Colors.black.withOpacity(0.3),
                      shape: BoxShape.circle
                    ),
                    child: const Padding(
                      padding: EdgeInsets.all(8.0),
                      child: Icon(
                        Icons.reply,
                        color: Colors.white,
                      )
                    )
                  ),
                ),
              ),
            );
          },
        );
      },
      isEelevated: false,
      key: ValueKey("message;" + item.toString()),
      child: ChatBubble(
        message: item,
        sentBySelf: item.sent,
        start: start,
        end: end,
        between: between,
        maxWidth: maxWidth,
      )
    );
  }

  void _block(ConversationState state, BuildContext context) {
    final jid = state.conversation!.jid;

    showConfirmationDialog(
      "Block $jid?",
      "Are you sure you want to block $jid? You won't receive messages from them until you unblock them.",
      context,
      () {
        context.read<ConversationBloc>().add(JidBlockedEvent(jid));
        Navigator.of(context).pop();
      }
    );
  }
  
  /// Render a widget that allows the user to either block the user or add them to their
  /// roster
  Widget _renderNotInRosterWidget(ConversationState state, BuildContext context) {
    return Container(
      color: Colors.black38,
      child: SizedBox(
        height: 64,
        child: Row(
          mainAxisAlignment: MainAxisAlignment.spaceBetween,
          children: [
            Expanded(
              child: TextButton(
                child: const Text("Add to contacts"),
                onPressed: () {
                  final jid = state.conversation!.jid;
                  showConfirmationDialog(
                    "Add $jid to your contacts?",
                    "Are you sure you want to add $jid to your conacts?",
                    context,
                    () {
                      // TODO: Maybe show a progress indicator
                      // TODO: Have the page update its state once the addition is done
                      context.read<ConversationBloc>().add(
                        JidAddedEvent(jid)
                      );
                      Navigator.of(context).pop();
                    }
                  );
                }
              )
            ),
            Expanded(
              child: TextButton(
                child: const Text("Block"),
                onPressed: () => _block(state, context)
              )
            )
          ]
        )
      )
    );
  }
  
  @override
  Widget build(BuildContext context) {
    double maxWidth = MediaQuery.of(context).size.width * 0.6;
    
    return WillPopScope(
      onWillPop: () async {
        GetIt.I.get<ConversationBloc>().add(CurrentConversationResetEvent());
        return true;
      },
      child: Stack(
        children: [
          Positioned(
            left: 0,
            right: 0,
            bottom: 0,
            child: BlocBuilder<ConversationBloc, ConversationState>(
              buildWhen: (prev, next) => prev.backgroundPath != next.backgroundPath,
              builder: (context, state) {
                final query = MediaQuery.of(context);
                return Image.file(
                  File(state.backgroundPath),
                  fit: BoxFit.cover,
                  width: query.size.width,
                  height: query.size.height - query.padding.top
                );
              }
            )
          ),
          Positioned(
            left: 0,
            right: 0,
            top: 0,
            bottom: 0,
            child: Scaffold(
              // TODO: Maybe replace the scaffold itself to prevent transparency
              backgroundColor: const Color.fromRGBO(0, 0, 0, 0),
              appBar: BorderlessTopbar(_ConversationTopbarWidget()),
              body: Column(
                crossAxisAlignment: CrossAxisAlignment.start,
                children: [
                  BlocBuilder<ConversationBloc, ConversationState>(
                    buildWhen: (prev, next) => prev.conversation?.inRoster != next.conversation?.inRoster,
                    builder: (context, state) {
                      if (state.conversation!.inRoster) return Container();

                      return _renderNotInRosterWidget(state, context);
                    }
                  ),

                  BlocBuilder<ConversationBloc, ConversationState>(
                    buildWhen: (prev, next) => prev.messages != next.messages,
                    builder: (context, state) => Expanded(
                      child: ListView.builder(
                        reverse: true,
                        itemCount: state.messages.length,
                        itemBuilder: (context, index) => _renderBubble(state, context, index, maxWidth),
                        shrinkWrap: true
                      )
                    )
                  ),

                  _ConversationBottomRow(_controller, _isSpeedDialOpen)
                ]
              )
            )
          ),
        ]
      )
    );
  }
}
